/* Configuration file parser
 *
 * Copyright (c) 2002, Martin Hedenfalk <mhe@home.se>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/* $Id: libcfg.h,v 1.7 2002/04/17 13:44:56 mhe Exp $
 */

#ifndef _cfg_h_
#define _cfg_h_

#include <stdio.h>
#include <stdarg.h>

typedef enum cfg_type_t cfg_type_t;
enum cfg_type_t {
	CFGT_NONE,
	CFGT_INT,
	CFGT_STR,
	CFGT_BOOL,
	CFGT_SEC
};

/* Flags
 */
#define CFGF_MULTI 1    /* option may be specified multiple times */
#define CFGF_LIST 2
#define CFGF_NOCASE 4
#define CFGF_TITLE 8

#define is_set(f, x) ((f & x) == f)

typedef union cfgval_t cfgval_t;
typedef struct cfgarray_t cfgarray_t;
typedef struct cfgopt_t cfgopt_t;
typedef struct cfg_t cfg_t;
typedef int (*cfg_callback_t)(cfg_t *cfg, const char *opt, const char *value);

struct cfgarray_t {
	int size;
	cfgval_t *val;
};

struct cfg_t {
	int flags;
	char *name;
	cfgopt_t *opts;
	char *title;

	const char *filename;
	int line;
};

union cfgval_t {
	int i;
	char *s;
	cfg_t sec;
	cfgarray_t a;
};

struct cfgopt_t {
	char *name;
	cfg_type_t type;
	cfg_callback_t *callback;
	cfgval_t val;
	int flags;
	cfgopt_t *subopts;
	void *def;
};

#define CFG_STR(name, def, flags)         {name, CFGT_STR, 0, {0}, flags, 0, (void *)def}
#define CFG_INT(name, def, flags)         {name, CFGT_INT, 0, {0}, flags, 0, (void *)def}
#define CFG_BOOL(name, def, flags)        {name, CFGT_BOOL, 0, {0}, flags, 0, (void *)def}
#define CFG_SEC(name, opts, flags)   {name, CFGT_SEC, 0, {0}, flags, opts, 0}
#define CFG_END()                    {0, CFGT_NONE, 0, {0}, 0, 0, 0}

typedef union {
	int i;
	char *s;
} YYSTYPE;
extern YYSTYPE yylval;

typedef void (*cfg_errfunc_t)(cfg_t *cfg, const char *fmt, va_list ap);

extern FILE *yyin;
extern int yyparse();

cfg_t *cfg_init(cfgopt_t *opts, int flags);
int cfg_parse(cfg_t *cfg, const char *filename);
void cfg_free(cfg_t *cfg);
cfg_errfunc_t cfg_set_error_function(cfg_errfunc_t errfunc);
void cfg_error(cfg_t *cfg, const char *fmt, ...);
cfgopt_t *cfg_addopt(cfg_t *cfg, cfgopt_t *opt);

cfgval_t *cfg_setopt(cfg_t *cfg, const char *name, void *val);
#define cfg_setint(cfg, name, val) cfg_setopt(cfg, name, (void *)val)
#define cfg_setstr(cfg, name, val) cfg_setopt(cfg, name, (void *)val)
cfgopt_t *cfg_getopt(cfg_t *cfg, const char *name);

cfgval_t *cfg_getval(cfg_t *cfg, const char *name);
int cfg_getint(cfg_t *cfg, const char *name);
char *cfg_getstr(cfg_t *cfg, const char *name);

cfg_t *cfg_tgetsec(cfg_t *cfg, const char *secname, const char *title);

cfgval_t *cfg_sgetval(cfg_t *cfg, const char *sec, const char *name);
int cfg_sgetint(cfg_t *cfg, const char *sec, const char *name);
char *cfg_sgetstr(cfg_t *cfg, const char *sec, const char *name);

cfgval_t *cfg_stgetval(cfg_t *cfg, const char *sec, const char *title, const char *name);
int cfg_stgetint(cfg_t *cfg, const char *sec, const char *title, const char *name);
char *cfg_stgetstr(cfg_t *cfg, const char *sec, const char *title, const char *name);

cfgarray_t *cfg_itemize(const char *str);

#endif
