/* Open.Net Authentication Daemon
 * ARP ping probe library
 *
 * Written by Martin Hedenfalk <mhe@it.kth.se>
 *
 * Copyright (c) 2002, Royal Institute of Technology, Stockholm, Sweden
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/* $Id: arping.c,v 1.8 2002/04/16 15:49:29 mhe Exp $
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdlib.h>
#include <unistd.h>
#include <libnet.h>
#include <pcap.h>

#include "common.h"

/* data passed to the packet callback function (called from
 * pcap_loop())
 */
struct packet_data_t {
	userdata_t *ud;
	int numrecvd; /* number of replies received */
};

typedef struct probe_arping_data_t probe_arping_data_t;
struct probe_arping_data_t {
	struct libnet_link_int *linkint;
	u_char *packet;
	pcap_t *pcap;
};

static void packet_callback(struct packet_data_t *pd,
							struct pcap_pkthdr *h,
							u_char *packet)
{
	struct libnet_ethernet_hdr *eth = (struct libnet_ethernet_hdr *)packet;
	struct libnet_arp_hdr *arp = (struct libnet_arp_hdr *)
		((char *)eth + sizeof(struct libnet_ethernet_hdr));

	if((htons(arp->ar_op) == ARPOP_REPLY) &&
	   (htons(arp->ar_pro) == ETHERTYPE_IP) &&
	   (htons(arp->ar_hrd) == ARPHRD_ETHER))
	{
		/*
		 * check if we got a response for the right IP address
		 */
		if(memcmp(arp->ar_spa, &pd->ud->ip.s_addr, 4) == 0)
		{
			/* Compare the source MAC address from the one passed in
			 * the userdata_t parameter.
			 */

#if 0
			logmsg(LOG_DEBUG, "ARP: %d.%d.%d.%d == %s?",
				   arp->ar_spa[0], arp->ar_spa[1],
				   arp->ar_spa[2], arp->ar_spa[3],
				   inet_ntoa(pd->ud->ip));
#endif

			if(memcmp(eth->ether_shost, pd->ud->mac, ETHER_ADDR_LEN) != 0) {
				char tmp[18];
				strcpy(tmp, strmac(pd->ud->mac));
				logmsg(LOG_ERR,
					   "got spoofed arp reply: %s, expected %s",
					   strmac(eth->ether_shost), tmp);
			} else
				pd->numrecvd++;
		}
		/* else we got an ARP reply from another IP (not for us) */
	}
}

void probe_cleanup(userdata_t *ud)
{
	if(ud && ud->probe_data) {
		probe_arping_data_t *p = (probe_arping_data_t *)ud->probe_data;
		if(p->pcap)
			pcap_close(p->pcap);
		if(p->packet)
			libnet_destroy_packet(&p->packet);
		if(p->linkint)
			libnet_close_link_interface(p->linkint);
		free(ud->probe_data);
		ud->probe_data = 0;
	}
}

int probe_user(userdata_t *ud, int wait_response)
{
	u_long myip = 0;
	char errbuf[LIBNET_ERRBUF_SIZE];
	char *ifname;
	struct packet_data_t pd;
	struct ether_addr *mymac;
	u_char eth_target[ETHER_ADDR_LEN] = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};
	u_char eth_source[ETHER_ADDR_LEN];
	pcap_t *pcap;
	u_char *packet;
	struct bpf_program filter;
	struct libnet_link_int *linkint;

	ifname = cfg_stgetstr(cfg, "domain", ud->domain->name, "probe-interface");

	logmsg(LOG_DEBUG, "arping %s on interface %s",
		   strmac(ud->mac), ifname);

	if(!(linkint = libnet_open_link_interface(ifname, errbuf))) {
		logmsg(LOG_ERR, "libnet_open_link_interface(): %s", errbuf);
		return -1;
	}
	ud->probe_data = (void *)malloc(sizeof(probe_arping_data_t));
	((probe_arping_data_t *)ud->probe_data)->linkint = linkint;
	((probe_arping_data_t *)ud->probe_data)->packet = 0;
	((probe_arping_data_t *)ud->probe_data)->pcap = 0;

	if(!(mymac = libnet_get_hwaddr(linkint, (u_char*)ifname, errbuf))) {
		logmsg(LOG_ERR, "libnet_get_hwaddr(): %s", errbuf);
		libnet_close_link_interface(linkint);
		return -1;
	}
	memcpy(eth_source, mymac->ether_addr_octet, ETHER_ADDR_LEN);

	if(!(myip = htonl(libnet_get_ipaddr(linkint, (u_char *)ifname, errbuf)))) {
		logmsg(LOG_ERR, "libnet_get_ipaddr(): %s", errbuf);
		libnet_close_link_interface(linkint);
		return -1;
	}

	if(libnet_init_packet(LIBNET_ETH_H + LIBNET_ARP_H, &packet) == -1) {
		logmsg(LOG_ERR, "libnet_init_packet(): error");
		libnet_close_link_interface(linkint);
		return -1;
	}
	((probe_arping_data_t *)ud->probe_data)->packet = packet;

	if(libnet_build_ethernet(eth_target, /* target MAC: broadcast */
							 eth_source,
							 ETHERTYPE_ARP,
							 NULL,
							 0,
							 packet) == -1)
	{
		logmsg(LOG_ERR, "libnet_build_ethernet(): error");
		libnet_close_link_interface(linkint);
		return -1;
	}

	if(libnet_build_arp(ARPHRD_ETHER, ETHERTYPE_IP, 6, 4, ARPOP_REQUEST,
						eth_source, (u_char *)&myip,
						eth_target, (u_char *)&ud->ip,
						NULL, 0, packet + LIBNET_ETH_H) == -1)
	{
		logmsg(LOG_ERR, "libnet_build_arp(): error");
		libnet_close_link_interface(linkint);
		return -1;
	}

	if(!(pcap = pcap_open_live(ifname, 100, 0, 10, errbuf))) {
		logmsg(LOG_ERR, "pcap_open_live(): %s", errbuf);
		libnet_close_link_interface(linkint);
		return -1;
	}
	((probe_arping_data_t *)ud->probe_data)->pcap = pcap;

	if(pcap_compile(pcap, &filter, "arp", 0, -1) == -1) {
		logmsg(LOG_ERR, "pcap_compile(): error");
		libnet_close_link_interface(linkint);
		return -1;
	}

	if(pcap_setfilter(pcap, &filter) == -1) {
		logmsg(LOG_ERR, "pcap_setfilter(): error");
		libnet_close_link_interface(linkint);
		return -1;
	}

#if 0
	{
		struct libnet_ethernet_hdr *eth = (struct libnet_ethernet_hdr *)packet;
		struct libnet_arp_hdr *arp = (struct libnet_arp_hdr *)
			((char *)eth + sizeof(struct libnet_ethernet_hdr));

		logmsg(LOG_DEBUG, "arping %s for ip %d.%d.%d.%d",
			   strmac(arp->ar_tha),
			   arp->ar_tpa[0], arp->ar_tpa[1], arp->ar_tpa[2], arp->ar_tpa[3]);
	}
#endif

	if(libnet_write_link_layer(linkint,
							   (u_char *)ifname,
							   (u_char *)packet,
							   LIBNET_ARP_H + LIBNET_ETH_H) == -1)
	{
		logmsg(LOG_ERR, "libnet_write_link_layer(): error");
		libnet_close_link_interface(linkint);
		return -1;
	}

	if(wait_response) {
		pd.ud = ud;
		pd.numrecvd = 0;
		while(pd.numrecvd == 0) {
			if(pcap_loop(pcap, 1,
						 (pcap_handler)packet_callback, (char *)&pd) == -1)
				logmsg(LOG_ERR, "pcap_loop(): error");
		}
	}
	libnet_destroy_packet(&packet);
	libnet_close_link_interface(linkint);
	free(ud->probe_data);
	ud->probe_data = 0;
	return pd.numrecvd > 0 ? 0 : -1;
}
