/* Open.Net Authentication Daemon
 * Functions to control the firewalls
 *
 * Written by Martin Hedenfalk <mhe@it.kth.se>
 *
 * Copyright (c) 2002, Royal Institute of Technology, Stockholm, Sweden
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "fwssl.h"

static void fw_check_connection(domain_t *domain)
{
	cfg_t *dsec = cfg_tgetsec(cfg, "domain", domain->name);

	if(domain->sslout == 0) {
		SSL_CTX *ctx = init_ctx(cfg_getstr(dsec, "certificate-file"),
								cfg_getstr(dsec, "key-file"),
								cfg_getstr(dsec, "fwcd-certificate-file"),
								1);
		if(ctx) {
			domain->sslout = init_ssl_client(
				ctx,
				cfg_getstr(dsec, "fwcd-host"),
				cfg_getint(dsec, "fwcd-port"));
		}
	}
}

static fwreply_t *fw_send_request(domain_t *domain, fwreq_t *req)
{
	static fwreply_t reply;
	int ret;

	reply.status = REQ_STATUS_FWCONNFAIL;
	reply.nstats = 0;

	/* FIXME: When an SSL_write()/SSL_read() operation has to be
	 * repeated because of SSL_ERROR_WANT_READ or
	 * SSL_ERROR_WANT_WRITE, it must be repeated with the same
	 * arguments.
	 */
	ret = SSL_write(domain->sslout, req, sizeof(fwreq_t));
	switch(check_ssl_io(ret, domain->sslout, sizeof(fwreq_t))) {
		case 0:
			/* success */
			break;
		case 1:
			/* reconnect and retry */
			SSL_free(domain->sslout);
			domain->sslout = 0;
			fw_check_connection(domain);
			if(domain->sslout == 0)
				/* reconnection failed */
				return &reply;
			return fw_send_request(domain, req);
			/* FIXME: can we end up in a infinite recursion loop here?
			 */
			break;
		case 2:
			/* give up */
			return &reply;
			break;
	}

	ret = SSL_read(domain->sslout, &reply, sizeof(fwreply_t));
	switch(check_ssl_io(ret, domain->sslout, sizeof(fwreply_t))) {
		case 0:
			/* success */
			break;
		case 1:
			/* reconnect and retry */
			SSL_free(domain->sslout);
			domain->sslout = 0;
			fw_check_connection(domain);
			if(domain->sslout == 0)
				/* reconnection failed */
				return &reply;
			return fw_send_request(domain, req);
			/* FIXME: can we end up in a infinite recursion loop here?
			 */
			break;
		case 2:
			/* give up */
			return &reply;
			break;
	}

	return &reply;
}

req_status_t fw_open(userdata_t *ud)
{
	req_status_t status = REQ_STATUS_OK;

	if(ud->domain->fwtype == FW_LOCAL) {
		status = fw_open_local(cfg_stgetstr(cfg, "domain", ud->domain->name,
											"firewall-program"),
							   &ud->ip, ud->mac);
	} else {
		/* Connect to firewall-host using SSL and send an OPEN
		 * command. Keep the connection open for future use.
		 */
		fwreq_t fwreq;
		fwreply_t *fwreply;

		fw_check_connection(ud->domain);
		if(ud->domain->sslout == 0) {
			/* unable to connect to firewall control daemon */
			status = REQ_STATUS_FWCONNFAIL;
		} else {
			fwreq.cmd = FWCMD_OPEN;
			memcpy(&fwreq.ip, &ud->ip, sizeof(struct in_addr));
			memcpy(fwreq.mac, ud->mac, 6);
			fwreply = fw_send_request(ud->domain, &fwreq);
			status = fwreply->status;
		}
	}

	if(status == REQ_STATUS_OK)
		gettimeofday(&ud->logintime, 0);

	return status;
}

req_status_t fw_close(userdata_t *ud)
{
	req_status_t status = REQ_STATUS_OK;

	if(ud->domain->fwtype == FW_LOCAL) {
		status = fw_close_local(cfg_stgetstr(cfg, "domain", ud->domain->name,
											 "firewall-program"),
								&ud->ip, ud->mac);
	} else {
		/* Connect to firewall-host using SSL and send an OPEN
		 * command. Keep the connection open for future use.
		 */
		fwreq_t fwreq;
		fwreply_t *fwreply;

		fw_check_connection(ud->domain);
		if(ud->domain->sslout == 0) {
			/* unable to connect to firewall control daemon */
			status = REQ_STATUS_FWCONNFAIL;
		} else {
			fwreq.cmd = FWCMD_CLOSE;
			memcpy(&fwreq.ip, &ud->ip, sizeof(struct in_addr));
			memcpy(fwreq.mac, ud->mac, 6);
			fwreply = fw_send_request(ud->domain, &fwreq);
			status = fwreply->status;
		}
	}
	return status;
}

void fw_reset(domain_t *domain)
{
	logmsg(LOG_INFO, "resetting firewall on domain %s", domain->name);

	if(domain->fwtype == FW_LOCAL) {
		fw_reset_local(cfg_stgetstr(cfg, "domain", domain->name,
									"firewall-program"));
	} else {
		fwreq_t fwreq;
		fwreply_t *fwreply;

		fw_check_connection(domain);

		if(domain->sslout == 0) {
			/* unable to connect to firewall control daemon */
		} else {
			fwreq.cmd = FWCMD_RESET;
			fwreply = fw_send_request(domain, &fwreq);
		}
	}
	/* return value is ignored
	 */
}
