#include <stdlib.h>
#include "llhash.h"

/*
 *  llhinit: Initialize the hash table
 *
 *  Parameters:
 *
 *    t           Hash table (duh!)
 */
void
llhinit(LLHTABLE t)
{
	int i = LLHSLOTS;

	while (i --) {
		t[i] = NULL;
	}
}

/*
 *  llhlookup: Access an item
 *
 *  Parameters:
 *
 *    t           Hash table
 *
 *    f           Pointer to hash function
 *
 *    key         Value uniquely identifying the item to be looked for
 *
 *  Return values:
 *
 *    non-NULL    The item in question was found; the return value is
 *                a pointer to the corresponding record in the hash table.
 *
 *    NULL        The item was not found in the hash table. Either it just
 *                isn't in there, or the hash function returned an error.
 */
LLHITEM *
llhlookup(LLHTABLE t, LLHFUNC *f, int key)
{
	int slot;
	struct item *tmp;

	if ((slot = (*f)(key) % LLHSLOTS) < 0) {
		return NULL;
	};
	tmp = t[slot];
	while (tmp) {
		if (LLHKEY(tmp->d) == key) {
			return &tmp->d;
		}
		tmp = tmp->next;
	}
	return NULL;
}

/*
 *  llhinsert: Insert an item into the hash table
 *
 *  Parameters:
 *
 *    t           Hash table
 *
 *    f           Pointer to hash function
 *
 *    d           Item to be inserted
 *
 *  Return values:
 *
 *    zero        The item was inserted successfully
 *
 *    negative    The item could not be inserted. The hash table
 *                already contains an item with the same key.
 *
 *    nonzero     The item could not be inserted. Either the function could
 *                not allocate the amount of memory necessary to store it,
 *                or the hash function returned an error.
 *
 *  Note:
 *
 *    If you know for sure that key values are in fact unique identifiers,
 *    that is, that the calling functions will never try to make the hash
 *    table contain two items with the same key at the same time, you can
 *    speed up the function considerably by deleting the first statement.
 */
int
llhinsert(LLHTABLE t, LLHFUNC *f, LLHITEM d)
{
	int slot;
	struct item *tmp;

	/* delete this line to insert items in constant time: */
	if (llhlookup(t, f, LLHKEY(d))) {
		logmsg(LOG_ERR, "key %u already exists", LLHKEY(d));
		return -1;
	}

	if ((slot = (*f)(LLHKEY(d)) % LLHSLOTS) < 0 
		|| (tmp = malloc(sizeof(struct item))) == NULL)
	{
		logmsg(LOG_ERR, "no slot or memory error");
		return 1;
	}
	tmp->d = d;
	tmp->next = t[slot];
	t[slot] = tmp;
	return 0;
}

/*
 *  llhremove: Remove an item from the hash table
 *
 *  Parameters:
 *
 *    t           Hash table
 *
 *    f           Pointer to hash function
 *
 *    key         Value uniquely identifying the item to be looked for
 *
 *  Return values:
 *
 *    nonzero     The item was removed successfully.
 *
 *    zero        The item could not be removed. Either it just wasn't
 *                found in the hash table, or the hash function returned
 *                an error.
 */
int
llhremove(LLHTABLE t, LLHFUNC *f, int key)
{
	int slot;
	struct item **tmp, *kill;

	if ((slot = (*f)(key) % LLHSLOTS) < 0) {
		return 0;
	}
	tmp = &t[slot];
	while (*tmp) {
		if (LLHKEY((*tmp)->d) == key) {
			kill = *tmp;
			*tmp = (*tmp)->next;
			free(kill);
			return 1;
		}
		tmp = &((*tmp)->next);
	}
	return 0;
}

void llhclear(LLHTABLE t)
{
	int i;

	for(i = 0; i < LLHSLOTS; ++i) {
		struct item **item = &t[i];
		struct item *kill;

		while(*item) {
			kill = *item;
			*item = (*item)->next;
			free(kill->d);
			free(kill);
			if(*item)
				item = &((*item)->next);
		}
	}
}

int llhsize(LLHTABLE t)
{
	int slot;
	int size = 0;

	for(slot = 0; slot < LLHSLOTS; ++slot) {
		struct item *i = t[slot];
		while(i) {
			++size;
			i = i->next;
		}
	}
	return size;
}
