/* Ping module
 *
 * Written by Fredrik Lilieblad <fli@it.kth.se>
 *
 * adapted for use in authd by Martin Hedenfalk <mhe@it.kth.se>
 */

/* $Id: ping.c,v 1.10 2002/04/10 13:26:25 mhe Exp $
 */

#include "common.h"
#include <netdb.h>
#include <libnet.h>

#define MAXPACKET (65536 - 60 - 8)	/* size of the largest packet */

int in_cksum(u_short * addr, int len)
{
	int nleft = len;
	u_short *w = addr;
	int sum = 0;
	u_short answer = 0;

	/*
	 * Our algorithm is simple, using a 32 bit accumulator (sum), we add
	 * sequential 16 bit words to it, and at the end, fold back all the
	 * carry bits from the top 16 bits into the lower 16 bits.
	 */
	while(nleft > 1) {
		sum += *w++;
		nleft -= 2;
	}

	/* mop up an odd byte, if necessary */
	if (nleft == 1) {
		*(u_char *) (&answer) = *(u_char *) w;
		sum += answer;
	}

	/* add back carry outs from top 16 bits to low 16 bits */
	sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
	sum += (sum >> 16);	/* add carry */
	answer = ~sum;		/* truncate to 16 bits */
	return (answer);
}

void probe_cleanup(userdata_t *ud)
{
	int sockfd = (int)ud->probe_data;

	close(sockfd);
}

/* ping probe function
 *
 * Sends an ICMP_ECHO to the address in ud->ip and waits for an
 * ICMP_ECHOREPLY
 */
int probe_user(userdata_t *ud, int wait_response)
{
	int sockfd;
	int pid;
	int protocol;

	struct sockaddr srv_addr;
	struct sockaddr from;
	struct sockaddr_in *to;
	int fromlen;

	struct protoent *proto;
	struct libnet_icmp_hdr *icp;
	struct libnet_ip_hdr  *ip;
	
	/* outgoing packet
	 */
	int cc, datalen, hlen;
	u_char outpackhdr[MAXPACKET];
	u_char *outpack = outpackhdr + sizeof(struct libnet_ip_hdr);

	/* incoming packet
	 */
	u_char inpackhdr[MAXPACKET];
	u_char *inpack = inpackhdr;

	pid = getpid();

	/* reserve space for ip (64) and time struct (8)
	 */
	datalen = 64 - 8;

	/* outgoing ICMP header
	 */
	icp = (struct libnet_icmp_hdr *) outpack;
	icp->icmp_type = ICMP_ECHO;
	icp->icmp_code = 23;
	icp->icmp_sum = 0;
	icp->hun.echo.seq = htons(23);	/* what else? */
	icp->hun.echo.id = pid;

	/* add time struct
	 */
	gettimeofday((struct timeval *) &outpack[8],
				 (struct timezone *) NULL);

	/* add for the time
	 */
	cc = datalen + 8;

	/* calculate new checksum
	 */
	icp->icmp_sum = in_cksum((u_short *) icp, cc);

	if(!(proto = getprotobyname("icmp"))) {
#if defined(IPPROTO_ICMP)
		protocol = IPPROTO_ICMP;
#else
		logmsg(LOG_ERR, "ping: can't get ICMP protocol\n");
		return -1;
#endif
	} else
		protocol = proto->p_proto;

	/* open a raw socket
	 */
	if ((sockfd = socket(AF_INET, SOCK_RAW, protocol)) == -1) {
		fprintf(stderr, "pinger: Can't open socket");
		return -1;
	}

	ud->probe_data = (void *)sockfd;

	/* compose a target address
	 */
	memset(&srv_addr, 0, sizeof(struct sockaddr));
	to = (struct sockaddr_in *) &srv_addr;
	to->sin_family = AF_INET;
	memcpy(&(to->sin_addr), (struct in_addr *)&ud->ip,
	       sizeof(struct in_addr));

	/* ...and send the packet
	 */
	if(sendto
	   (sockfd, (char *) outpack, cc, wait_response ? 0 : MSG_DONTWAIT,
		(struct sockaddr *) &srv_addr,
		sizeof(struct sockaddr)) == -1)
	{
		logmsg(LOG_ERR, "ping: can't send packet: %s", strerror(errno));
		close(sockfd);
		return -1;
	}

	if(wait_response) {
		/* loop, waiting for replies
		 */
		do {
			recvfrom(sockfd, (char *) inpack, MAXPACKET, 0,
					 (struct sockaddr *) &from, &fromlen);
			ip = (struct libnet_ip_hdr *) inpack;
			hlen = ip->ip_hl * 4;
			icp = (struct libnet_icmp_hdr *) (inpack + hlen);
		} while((icp->icmp_type != ICMP_ECHOREPLY || icp->hun.echo.id != pid
				 || ntohs(icp->hun.echo.seq) != 23));
	}

	shutdown(sockfd, 2);
	close(sockfd);

	return 0;
}
