/* Open.Net Authentication Daemon
 * Spawning of a child process
 *
 * Written by Martin Hedenfalk <mhe@it.kth.se>
 *
 * Copyright (c) 2002, Royal Institute of Technology, Stockholm, Sweden
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/* $Id: spawn.c,v 1.4 2002/04/06 21:27:11 mhe Exp $
 */

#include "common.h"

static pthread_key_t pid_key, sent_key;

#define SEND_TERM 1
#define SEND_KILL 2

static void gazonk(int signo)
{
	logmsg(LOG_ERR, "SHOULD NOT HAPPEN!");
}

static void cancel_process(int signo)
{
	int s = (int)pthread_getspecific(sent_key);
	pid_t pid = (pid_t)pthread_getspecific(pid_key);

	if(s == 0) {
		logmsg(LOG_ERR, "Failed reading sent_key");
		return;
	}
	if(pid == 0) {
		logmsg(LOG_ERR, "Failed reading pid_key");
		return;
	}

	if(s == SEND_KILL) {
		logmsg(LOG_ERR, "timeout waiting for process, sending SIGKILL to %d",
			   pid);
		kill(pid, SIGKILL);
		alarm(0);
		signal(SIGALRM, gazonk);
	} else {
		logmsg(LOG_ERR, "timeout waiting for process, sending SIGTERM to %d",
			   pid);
		kill(pid, SIGTERM);
		if(pthread_setspecific(sent_key, (void *)SEND_KILL) != 0)
			logmsg(LOG_ERR, "failed updating sent_key");
		signal(SIGALRM, cancel_process);
		alarm(cfg_getint(cfg, "firewall-hard-timeout"));
	}
}

/* Fork and exec a new process, then wait for completion and return
 * the exit status. Soft and hard timeout can be set in the
 * configuration file with firewall-soft-timeout/firewall-hard-timeout.
 */
int spawn(char * const argv[])
{
	int status;
	pid_t spawn_pid;

    spawn_pid = fork();

	pthread_key_create(&sent_key, NULL);
	if(pthread_setspecific(sent_key, (void *)SEND_TERM) != 0)
		logmsg(LOG_ERR, "failed resetting sent_key");
	pthread_key_create(&pid_key, NULL);
	if(pthread_setspecific(pid_key, (void *)spawn_pid) != 0)
		logmsg(LOG_ERR, "failed setting pid_key");

    if(spawn_pid == 0) {
        if(execv(argv[0], argv) < 0) {
			logmsg(LOG_ERR, "unable to exec %s: %s", argv[0], strerror(errno));
            exit(5);
        }
		exit(17);
    } else if(spawn_pid < 0) {
		logmsg(LOG_ERR, "failed to fork: %s", strerror(errno));
        return -1;
	}

	logmsg(LOG_DEBUG, "spawned %s[%d]", argv[0], spawn_pid);
	signal(SIGALRM, cancel_process);
	alarm(cfg_getint(cfg, "firewall-soft-timeout"));
	waitpid(spawn_pid, &status, 0);
	logmsg(LOG_DEBUG, "%s[%d] returned status %d", argv[0], spawn_pid, status);
	alarm(0);
	signal(SIGALRM, gazonk);
    return status;
}
